"""Top-level package for synthesisai."""

__author__ = """Synthesis AI"""
__email__ = 'anish@synthesis.ai'
__version__ = '0.1.0'

from .dataset import Grouping, SaiDataset
from .modality import Modality
